{
    "volumeAttachment": {
        "device": "/dev/sdd",
        "id": "a26887c6-c47b-4654-abb5-dfadf7d3f803",
        "serverId": "%(uuid)s",
        "volumeId": "a26887c6-c47b-4654-abb5-dfadf7d3f803"
    }
}
