{
    "floating_ips": [
        {
            "fixed_ip": null,
            "id": 1,
            "instance_id": null,
            "ip": "10.10.10.1",
            "pool": "nova"
        },
        {
            "fixed_ip": null,
            "id": 2,
            "instance_id": null,
            "ip": "10.10.10.2",
            "pool": "nova"
        }
    ]
}

