{
    "security_group_default_rule": {
        "from_port": 80,
        "id": 1,
        "ip_protocol": "TCP",
        "ip_range":{
            "cidr": "10.10.10.0/24"
        },
        "to_port": 80
    }
}