{
    "networks": [
        {
            "bridge": "br100",
            "bridge_interface": "eth0",
            "broadcast": "%(ip)s",
            "cidr": "10.0.0.0/29",
            "cidr_v6": null,
            "created_at": "%(strtime)s",
            "deleted": false,
            "deleted_at": null,
            "dhcp_start": "%(ip)s",
            "dns1": null,
            "dns2": null,
            "gateway": "%(ip)s",
            "gateway_v6": null,
            "host": "nsokolov-desktop",
            "id": "%(id)s",
            "injected": false,
            "label": "mynet_0",
            "multi_host": false,
            "netmask": "%(ip)s",
            "netmask_v6": null,
            "priority": null,
            "project_id": "6133f8b603924f45bc0c9e21f6df12fa",
            "rxtx_base": null,
            "updated_at": "%(strtime)s",
            "vlan": 100,
            "vpn_private_address": "%(ip)s",
            "vpn_public_address": "%(ip)s",
            "vpn_public_port": 1000,
            "mtu": null,
            "dhcp_server": "%(ip)s",
            "enable_dhcp": true,
            "share_address": false
        },
        {
            "bridge": "br101",
            "bridge_interface": "eth0",
            "broadcast": "%(ip)s",
            "cidr": "10.0.0.10/29",
            "cidr_v6": null,
            "created_at": "%(strtime)s",
            "deleted": false,
            "deleted_at": null,
            "dhcp_start": "%(ip)s",
            "dns1": null,
            "dns2": null,
            "gateway": "%(ip)s",
            "gateway_v6": null,
            "host": null,
            "id": "%(id)s",
            "injected": false,
            "label": "mynet_1",
            "multi_host": false,
            "netmask": "%(ip)s",
            "netmask_v6": null,
            "priority": null,
            "project_id": null,
            "rxtx_base": null,
            "updated_at": null,
            "vlan": 101,
            "vpn_private_address": "%(ip)s",
            "vpn_public_address": null,
            "vpn_public_port": 1001,
            "mtu": null,
            "dhcp_server": "%(ip)s",
            "enable_dhcp": true,
            "share_address": false
        }
    ]
}
