# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

common_user_data = {
    'type': 'string',
    'format': 'base64',
    'maxLength': 65535
}


server_create = {
    'user_data': common_user_data
}


server_create_v20 = {
    'user_data': {
        'oneOf': [
            common_user_data,
            {'type': 'null'},
        ],
    },
}
