{
    "domain_entries": [
        {
            "availability_zone": null,
            "domain": "%(domain)s",
            "project": "%(project)s",
            "scope": "%(scope)s"
        }
    ]
}
