{
    "floating_ip_info": [
        {
            "address": "10.10.10.1",
            "instance_uuid": null,
            "interface": "eth0",
            "pool": "nova",
            "fixed_ip": null,
            "project_id": null
        },
        {
            "address": "10.10.10.2",
            "instance_uuid": null,
            "interface": "eth0",
            "pool": "nova",
            "fixed_ip": null,
            "project_id": null
        },
        {
            "address": "10.10.10.3",
            "instance_uuid": null,
            "interface": "eth0",
            "pool": "nova",
            "fixed_ip": null,
            "project_id": null
        }
    ]
}
