{
    "instanceAction": {
            "action": "%(action)s",
            "instance_uuid": "%(instance_uuid)s",
            "request_id": "%(request_id)s",
            "user_id": "%(integer_id)s",
            "project_id": "%(integer_id)s",
            "start_time": "%(strtime)s",
            "message": "",
            "events": [
                {
                    "event": "%(event)s",
                    "start_time": "%(strtime)s",
                    "finish_time": "%(strtime)s",
                    "result": "%(result)s",
                    "traceback": ""
                },
                {
                    "event": "%(event)s",
                    "start_time": "%(strtime)s",
                    "finish_time": "%(strtime)s",
                    "result": "%(result)s",
                    "traceback": ""
                }
            ]
    }
}
