# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import MetaData, String, Table


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    table = Table('volume_usage_cache', meta, autoload=True)
    col_resource = getattr(table.c, 'user_id')
    # Match the maximum length of user_id in Keystone here instead of
    # assuming explicitly a single UUID length.
    col_resource.alter(type=String(64))
